import os
from typing import Type

import pytest
from typeguard import TypeCheckError

from grpc4bmi.bmi_client_apptainer import SUPPORTED_APPTAINER_VERSIONS, BmiClientApptainer, check_apptainer_version_string
from grpc4bmi.exceptions import ApptainerVersionException, DeadContainerException

class Test_check_apptainer_version_string:
    @pytest.mark.parametrize("test_input", [
        ('apptainer version 1.0.0-rc.2'),  
        ('apptainer version 1.0.0'),
        ('apptainer version 1.0.3'),
        ('apptainer version 1.1.0-rc.3'),
        ('apptainer version 1.1.2'),
        # From snellius cluster at SURF.
        ('apptainer version 1.1.5-2.el8'),
    ])
    def test_ok(self, test_input: str):
        result = check_apptainer_version_string(test_input)
        assert result

    @pytest.mark.parametrize("test_input,error_class,expected", [
        ('apptainer version 1.0.0-rc.1', ApptainerVersionException, SUPPORTED_APPTAINER_VERSIONS),
        ('apptainer version 0.1.0', ApptainerVersionException, SUPPORTED_APPTAINER_VERSIONS),
    ])
    def test_too_old(self, test_input: str, error_class: Type[ApptainerVersionException], expected: str):
        with pytest.raises(error_class, match=expected):
            check_apptainer_version_string(test_input)

IMAGE_NAME = "docker://ewatercycle/walrus-grpc4bmi:v0.2.0"

@pytest.fixture
def walrus_model(tmp_path, walrus_input):
    model = BmiClientApptainer(image=IMAGE_NAME, work_dir=str(tmp_path))
    yield model
    del model

class TestBmiClientApptainerHappyDays:
    def test_component_name(self, walrus_model):
        assert walrus_model.get_component_name() == 'WALRUS'

class TestBmiClientApptainerBadDays:
    def test_get_value_ref(self, walrus_model):
        with pytest.raises(NotImplementedError):
            walrus_model.get_value_ref('Q')

    def test_inputdir_absent(self, tmp_path):
        dirthatdoesnotexist = 'dirthatdoesnotexist'
        input_dir = tmp_path / dirthatdoesnotexist
        work_dir = tmp_path / 'work'
        work_dir.mkdir()
        with pytest.raises(NotADirectoryError, match=dirthatdoesnotexist):
            BmiClientApptainer(image=IMAGE_NAME, work_dir=str(work_dir), input_dirs=[str(input_dir)])

    def test_workdir_absent(self, tmp_path):
        dirthatdoesnotexist = 'dirthatdoesnotexist'
        work_dir = tmp_path / dirthatdoesnotexist
        with pytest.raises(NotADirectoryError, match=dirthatdoesnotexist):
            BmiClientApptainer(image=IMAGE_NAME, work_dir=str(work_dir))

    def test_same_inputdir_and_workdir(self, tmp_path):
        some_dir = str(tmp_path)
        match = 'Found work_dir equal to one of the input directories. Please drop that input dir.'
        with pytest.raises(ValueError, match=match):
            BmiClientApptainer(image=IMAGE_NAME, input_dirs=(some_dir,), work_dir=some_dir)

    def test_workdir_as_number(self):
        with pytest.raises(TypeCheckError, match='is not an instance of str'):
            BmiClientApptainer(image=IMAGE_NAME, work_dir=42)

    def test_inputdirs_as_str(self, tmp_path):
        some_dir = str(tmp_path)
        with pytest.raises(TypeError, match=' must be collections.abc.Iterable'):
            BmiClientApptainer(image=IMAGE_NAME, input_dirs='old type', work_dir=some_dir)

    def test_inputdirs_as_number(self, tmp_path):
        some_dir = str(tmp_path)
        with pytest.raises(TypeCheckError, match='is not an instance of collections.abc.Iterable'):
            BmiClientApptainer(image=IMAGE_NAME, input_dirs=42, work_dir=some_dir)



class TestRedirectOutput:
    EXPECTED = 'Hello from Docker!'

    @pytest.fixture(scope="module")
    def image(self):
        hello_image = 'docker://hello-world'
        # Cache image, first test does not use delay time to build image
        os.system(f'apptainer run {hello_image}')
        return hello_image

    def test_default(self, image, tmp_path, capfd):
        with pytest.raises(DeadContainerException) as excinf:
            BmiClientApptainer(image=image, work_dir=str(tmp_path), delay=2)

        assert self.EXPECTED not in capfd.readouterr().out
        assert self.EXPECTED in excinf.value.logs

    def test_devnull(self, image, tmp_path, capfd):
        with pytest.raises(DeadContainerException) as excinf:
            BmiClientApptainer(image=image, work_dir=str(tmp_path), capture_logs=False, delay=2)

        assert self.EXPECTED not in capfd.readouterr().out
        assert self.EXPECTED not in excinf.value.logs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc4bmi/bmi.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12grpc4bmi/bmi.proto\x12\x03\x62mi\"\x07\n\x05\x45mpty\"(\n\x11InitializeRequest\x12\x13\n\x0b\x63onfig_file\x18\x01 \x01(\t\"(\n\x18GetComponentNameResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\"$\n\x13GetVarNamesResponse\x12\r\n\x05names\x18\x01 \x03(\t\"%\n\x14GetTimeUnitsResponse\x12\r\n\x05units\x18\x01 \x01(\t\"\'\n\x13GetTimeStepResponse\x12\x10\n\x08interval\x18\x01 \x01(\x01\"\x1f\n\x0fGetTimeResponse\x12\x0c\n\x04time\x18\x01 \x01(\x01\"\x1d\n\rGetVarRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"%\n\x12GetVarGridResponse\x12\x0f\n\x07grid_id\x18\x01 \x01(\x05\"\"\n\x12GetVarTypeResponse\x12\x0c\n\x04type\x18\x01 \x01(\t\"&\n\x16GetVarItemSizeResponse\x12\x0c\n\x04size\x18\x01 \x01(\x03\"$\n\x13GetVarUnitsResponse\x12\r\n\x05units\x18\x01 \x01(\t\"&\n\x14GetVarNBytesResponse\x12\x0e\n\x06nbytes\x18\x01 \x01(\x03\"z\n\x16GetVarLocationResponse\x12\x36\n\x08location\x18\x01 \x01(\x0e\x32$.bmi.GetVarLocationResponse.Location\"(\n\x08Location\x12\x08\n\x04NODE\x10\x00\x12\x08\n\x04\x45\x44GE\x10\x01\x12\x08\n\x04\x46\x41\x43\x45\x10\x02\"%\n\x0fIntArrayMessage\x12\x12\n\x06values\x18\x01 \x03(\x03\x42\x02\x10\x01\"\'\n\x11\x46loatArrayMessage\x12\x12\n\x06values\x18\x01 \x03(\x02\x42\x02\x10\x01\"(\n\x12\x44oubleArrayMessage\x12\x12\n\x06values\x18\x01 \x03(\x01\x42\x02\x10\x01\"\xaa\x01\n\x10GetValueResponse\x12*\n\nvalues_int\x18\x01 \x01(\x0b\x32\x14.bmi.IntArrayMessageH\x00\x12.\n\x0cvalues_float\x18\x02 \x01(\x0b\x32\x16.bmi.FloatArrayMessageH\x00\x12\x30\n\rvalues_double\x18\x03 \x01(\x0b\x32\x17.bmi.DoubleArrayMessageH\x00\x42\x08\n\x06values\"=\n\x18GetValueAtIndicesRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x07indices\x18\x02 \x03(\x03\x42\x02\x10\x01\"\xb3\x01\n\x19GetValueAtIndicesResponse\x12*\n\nvalues_int\x18\x01 \x01(\x0b\x32\x14.bmi.IntArrayMessageH\x00\x12.\n\x0cvalues_float\x18\x02 \x01(\x0b\x32\x16.bmi.FloatArrayMessageH\x00\x12\x30\n\rvalues_double\x18\x03 \x01(\x0b\x32\x17.bmi.DoubleArrayMessageH\x00\x42\x08\n\x06values\"\xb7\x01\n\x0fSetValueRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\nvalues_int\x18\x02 \x01(\x0b\x32\x14.bmi.IntArrayMessageH\x00\x12.\n\x0cvalues_float\x18\x03 \x01(\x0b\x32\x16.bmi.FloatArrayMessageH\x00\x12\x30\n\rvalues_double\x18\x04 \x01(\x0b\x32\x17.bmi.DoubleArrayMessageH\x00\x42\x08\n\x06values\"/\n\x12SetValuePtrRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03ref\x18\x02 \x01(\x03\"\xd5\x01\n\x18SetValueAtIndicesRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x07indices\x18\x02 \x03(\x03\x42\x02\x10\x01\x12*\n\nvalues_int\x18\x03 \x01(\x0b\x32\x14.bmi.IntArrayMessageH\x00\x12.\n\x0cvalues_float\x18\x04 \x01(\x0b\x32\x16.bmi.FloatArrayMessageH\x00\x12\x30\n\rvalues_double\x18\x05 \x01(\x0b\x32\x17.bmi.DoubleArrayMessageH\x00\x42\x08\n\x06values\"\x1e\n\x0bGridRequest\x12\x0f\n\x07grid_id\x18\x01 \x01(\x03\"#\n\x13GetGridSizeResponse\x12\x0c\n\x04size\x18\x01 \x01(\x03\"#\n\x13GetGridRankResponse\x12\x0c\n\x04rank\x18\x01 \x01(\x03\"#\n\x13GetGridTypeResponse\x12\x0c\n\x04type\x18\x01 \x01(\t\")\n\x14GetGridShapeResponse\x12\x11\n\x05shape\x18\x01 \x03(\x03\x42\x02\x10\x01\"-\n\x16GetGridSpacingResponse\x12\x13\n\x07spacing\x18\x01 \x03(\x01\x42\x02\x10\x01\"+\n\x15GetGridOriginResponse\x12\x12\n\x06origin\x18\x01 \x03(\x01\x42\x02\x10\x01\"0\n\x15GetGridPointsResponse\x12\x17\n\x0b\x63oordinates\x18\x01 \x03(\x01\x42\x02\x10\x01\"!\n\x10GetCountResponse\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\"2\n\x18GetGridEdgeNodesResponse\x12\x16\n\nedge_nodes\x18\x01 \x03(\x03\x42\x02\x10\x01\"2\n\x18GetGridFaceEdgesResponse\x12\x16\n\nface_edges\x18\x01 \x03(\x03\x42\x02\x10\x01\"2\n\x18GetGridFaceNodesResponse\x12\x16\n\nface_nodes\x18\x01 \x03(\x03\x42\x02\x10\x01\"9\n\x1bGetGridNodesPerFaceResponse\x12\x1a\n\x0enodes_per_face\x18\x01 \x03(\x03\x42\x02\x10\x01\x32\x8a\x13\n\nBmiService\x12\x32\n\ninitialize\x12\x16.bmi.InitializeRequest\x1a\n.bmi.Empty\"\x00\x12\"\n\x06update\x12\n.bmi.Empty\x1a\n.bmi.Empty\"\x00\x12\x31\n\x0bupdateUntil\x12\x14.bmi.GetTimeResponse\x1a\n.bmi.Empty\"\x00\x12$\n\x08\x66inalize\x12\n.bmi.Empty\x1a\n.bmi.Empty\"\x00\x12?\n\x10getComponentName\x12\n.bmi.Empty\x1a\x1d.bmi.GetComponentNameResponse\"\x00\x12\x38\n\x11getInputItemCount\x12\n.bmi.Empty\x1a\x15.bmi.GetCountResponse\"\x00\x12\x39\n\x12getOutputItemCount\x12\n.bmi.Empty\x1a\x15.bmi.GetCountResponse\"\x00\x12:\n\x10getInputVarNames\x12\n.bmi.Empty\x1a\x18.bmi.GetVarNamesResponse\"\x00\x12;\n\x11getOutputVarNames\x12\n.bmi.Empty\x1a\x18.bmi.GetVarNamesResponse\"\x00\x12\x37\n\x0cgetTimeUnits\x12\n.bmi.Empty\x1a\x19.bmi.GetTimeUnitsResponse\"\x00\x12\x35\n\x0bgetTimeStep\x12\n.bmi.Empty\x1a\x18.bmi.GetTimeStepResponse\"\x00\x12\x34\n\x0egetCurrentTime\x12\n.bmi.Empty\x1a\x14.bmi.GetTimeResponse\"\x00\x12\x32\n\x0cgetStartTime\x12\n.bmi.Empty\x1a\x14.bmi.GetTimeResponse\"\x00\x12\x30\n\ngetEndTime\x12\n.bmi.Empty\x1a\x14.bmi.GetTimeResponse\"\x00\x12;\n\ngetVarGrid\x12\x12.bmi.GetVarRequest\x1a\x17.bmi.GetVarGridResponse\"\x00\x12;\n\ngetVarType\x12\x12.bmi.GetVarRequest\x1a\x17.bmi.GetVarTypeResponse\"\x00\x12\x43\n\x0egetVarItemSize\x12\x12.bmi.GetVarRequest\x1a\x1b.bmi.GetVarItemSizeResponse\"\x00\x12=\n\x0bgetVarUnits\x12\x12.bmi.GetVarRequest\x1a\x18.bmi.GetVarUnitsResponse\"\x00\x12?\n\x0cgetVarNBytes\x12\x12.bmi.GetVarRequest\x1a\x19.bmi.GetVarNBytesResponse\"\x00\x12\x43\n\x0egetVarLocation\x12\x12.bmi.GetVarRequest\x1a\x1b.bmi.GetVarLocationResponse\"\x00\x12\x37\n\x08getValue\x12\x12.bmi.GetVarRequest\x1a\x15.bmi.GetValueResponse\"\x00\x12T\n\x11getValueAtIndices\x12\x1d.bmi.GetValueAtIndicesRequest\x1a\x1e.bmi.GetValueAtIndicesResponse\"\x00\x12.\n\x08setValue\x12\x14.bmi.SetValueRequest\x1a\n.bmi.Empty\"\x00\x12@\n\x11setValueAtIndices\x12\x1d.bmi.SetValueAtIndicesRequest\x1a\n.bmi.Empty\"\x00\x12;\n\x0bgetGridSize\x12\x10.bmi.GridRequest\x1a\x18.bmi.GetGridSizeResponse\"\x00\x12;\n\x0bgetGridType\x12\x10.bmi.GridRequest\x1a\x18.bmi.GetGridTypeResponse\"\x00\x12;\n\x0bgetGridRank\x12\x10.bmi.GridRequest\x1a\x18.bmi.GetGridRankResponse\"\x00\x12=\n\x0cgetGridShape\x12\x10.bmi.GridRequest\x1a\x19.bmi.GetGridShapeResponse\"\x00\x12\x41\n\x0egetGridSpacing\x12\x10.bmi.GridRequest\x1a\x1b.bmi.GetGridSpacingResponse\"\x00\x12?\n\rgetGridOrigin\x12\x10.bmi.GridRequest\x1a\x1a.bmi.GetGridOriginResponse\"\x00\x12:\n\x08getGridX\x12\x10.bmi.GridRequest\x1a\x1a.bmi.GetGridPointsResponse\"\x00\x12:\n\x08getGridY\x12\x10.bmi.GridRequest\x1a\x1a.bmi.GetGridPointsResponse\"\x00\x12:\n\x08getGridZ\x12\x10.bmi.GridRequest\x1a\x1a.bmi.GetGridPointsResponse\"\x00\x12=\n\x10getGridNodeCount\x12\x10.bmi.GridRequest\x1a\x15.bmi.GetCountResponse\"\x00\x12=\n\x10getGridEdgeCount\x12\x10.bmi.GridRequest\x1a\x15.bmi.GetCountResponse\"\x00\x12=\n\x10getGridFaceCount\x12\x10.bmi.GridRequest\x1a\x15.bmi.GetCountResponse\"\x00\x12\x45\n\x10getGridEdgeNodes\x12\x10.bmi.GridRequest\x1a\x1d.bmi.GetGridEdgeNodesResponse\"\x00\x12\x45\n\x10getGridFaceNodes\x12\x10.bmi.GridRequest\x1a\x1d.bmi.GetGridFaceNodesResponse\"\x00\x12\x45\n\x10getGridFaceEdges\x12\x10.bmi.GridRequest\x1a\x1d.bmi.GetGridFaceEdgesResponse\"\x00\x12K\n\x13getGridNodesPerFace\x12\x10.bmi.GridRequest\x1a .bmi.GetGridNodesPerFaceResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc4bmi.bmi_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _INTARRAYMESSAGE.fields_by_name['values']._options = None
  _INTARRAYMESSAGE.fields_by_name['values']._serialized_options = b'\020\001'
  _FLOATARRAYMESSAGE.fields_by_name['values']._options = None
  _FLOATARRAYMESSAGE.fields_by_name['values']._serialized_options = b'\020\001'
  _DOUBLEARRAYMESSAGE.fields_by_name['values']._options = None
  _DOUBLEARRAYMESSAGE.fields_by_name['values']._serialized_options = b'\020\001'
  _GETVALUEATINDICESREQUEST.fields_by_name['indices']._options = None
  _GETVALUEATINDICESREQUEST.fields_by_name['indices']._serialized_options = b'\020\001'
  _SETVALUEATINDICESREQUEST.fields_by_name['indices']._options = None
  _SETVALUEATINDICESREQUEST.fields_by_name['indices']._serialized_options = b'\020\001'
  _GETGRIDSHAPERESPONSE.fields_by_name['shape']._options = None
  _GETGRIDSHAPERESPONSE.fields_by_name['shape']._serialized_options = b'\020\001'
  _GETGRIDSPACINGRESPONSE.fields_by_name['spacing']._options = None
  _GETGRIDSPACINGRESPONSE.fields_by_name['spacing']._serialized_options = b'\020\001'
  _GETGRIDORIGINRESPONSE.fields_by_name['origin']._options = None
  _GETGRIDORIGINRESPONSE.fields_by_name['origin']._serialized_options = b'\020\001'
  _GETGRIDPOINTSRESPONSE.fields_by_name['coordinates']._options = None
  _GETGRIDPOINTSRESPONSE.fields_by_name['coordinates']._serialized_options = b'\020\001'
  _GETGRIDEDGENODESRESPONSE.fields_by_name['edge_nodes']._options = None
  _GETGRIDEDGENODESRESPONSE.fields_by_name['edge_nodes']._serialized_options = b'\020\001'
  _GETGRIDFACEEDGESRESPONSE.fields_by_name['face_edges']._options = None
  _GETGRIDFACEEDGESRESPONSE.fields_by_name['face_edges']._serialized_options = b'\020\001'
  _GETGRIDFACENODESRESPONSE.fields_by_name['face_nodes']._options = None
  _GETGRIDFACENODESRESPONSE.fields_by_name['face_nodes']._serialized_options = b'\020\001'
  _GETGRIDNODESPERFACERESPONSE.fields_by_name['nodes_per_face']._options = None
  _GETGRIDNODESPERFACERESPONSE.fields_by_name['nodes_per_face']._serialized_options = b'\020\001'
  _EMPTY._serialized_start=27
  _EMPTY._serialized_end=34
  _INITIALIZEREQUEST._serialized_start=36
  _INITIALIZEREQUEST._serialized_end=76
  _GETCOMPONENTNAMERESPONSE._serialized_start=78
  _GETCOMPONENTNAMERESPONSE._serialized_end=118
  _GETVARNAMESRESPONSE._serialized_start=120
  _GETVARNAMESRESPONSE._serialized_end=156
  _GETTIMEUNITSRESPONSE._serialized_start=158
  _GETTIMEUNITSRESPONSE._serialized_end=195
  _GETTIMESTEPRESPONSE._serialized_start=197
  _GETTIMESTEPRESPONSE._serialized_end=236
  _GETTIMERESPONSE._serialized_start=238
  _GETTIMERESPONSE._serialized_end=269
  _GETVARREQUEST._serialized_start=271
  _GETVARREQUEST._serialized_end=300
  _GETVARGRIDRESPONSE._serialized_start=302
  _GETVARGRIDRESPONSE._serialized_end=339
  _GETVARTYPERESPONSE._serialized_start=341
  _GETVARTYPERESPONSE._serialized_end=375
  _GETVARITEMSIZERESPONSE._serialized_start=377
  _GETVARITEMSIZERESPONSE._serialized_end=415
  _GETVARUNITSRESPONSE._serialized_start=417
  _GETVARUNITSRESPONSE._serialized_end=453
  _GETVARNBYTESRESPONSE._serialized_start=455
  _GETVARNBYTESRESPONSE._serialized_end=493
  _GETVARLOCATIONRESPONSE._serialized_start=495
  _GETVARLOCATIONRESPONSE._serialized_end=617
  _GETVARLOCATIONRESPONSE_LOCATION._serialized_start=577
  _GETVARLOCATIONRESPONSE_LOCATION._serialized_end=617
  _INTARRAYMESSAGE._serialized_start=619
  _INTARRAYMESSAGE._serialized_end=656
  _FLOATARRAYMESSAGE._serialized_start=658
  _FLOATARRAYMESSAGE._serialized_end=697
  _DOUBLEARRAYMESSAGE._serialized_start=699
  _DOUBLEARRAYMESSAGE._serialized_end=739
  _GETVALUERESPONSE._serialized_start=742
  _GETVALUERESPONSE._serialized_end=912
  _GETVALUEATINDICESREQUEST._serialized_start=914
  _GETVALUEATINDICESREQUEST._serialized_end=975
  _GETVALUEATINDICESRESPONSE._serialized_start=978
  _GETVALUEATINDICESRESPONSE._serialized_end=1157
  _SETVALUEREQUEST._serialized_start=1160
  _SETVALUEREQUEST._serialized_end=1343
  _SETVALUEPTRREQUEST._serialized_start=1345
  _SETVALUEPTRREQUEST._serialized_end=1392
  _SETVALUEATINDICESREQUEST._serialized_start=1395
  _SETVALUEATINDICESREQUEST._serialized_end=1608
  _GRIDREQUEST._serialized_start=1610
  _GRIDREQUEST._serialized_end=1640
  _GETGRIDSIZERESPONSE._serialized_start=1642
  _GETGRIDSIZERESPONSE._serialized_end=1677
  _GETGRIDRANKRESPONSE._serialized_start=1679
  _GETGRIDRANKRESPONSE._serialized_end=1714
  _GETGRIDTYPERESPONSE._serialized_start=1716
  _GETGRIDTYPERESPONSE._serialized_end=1751
  _GETGRIDSHAPERESPONSE._serialized_start=1753
  _GETGRIDSHAPERESPONSE._serialized_end=1794
  _GETGRIDSPACINGRESPONSE._serialized_start=1796
  _GETGRIDSPACINGRESPONSE._serialized_end=1841
  _GETGRIDORIGINRESPONSE._serialized_start=1843
  _GETGRIDORIGINRESPONSE._serialized_end=1886
  _GETGRIDPOINTSRESPONSE._serialized_start=1888
  _GETGRIDPOINTSRESPONSE._serialized_end=1936
  _GETCOUNTRESPONSE._serialized_start=1938
  _GETCOUNTRESPONSE._serialized_end=1971
  _GETGRIDEDGENODESRESPONSE._serialized_start=1973
  _GETGRIDEDGENODESRESPONSE._serialized_end=2023
  _GETGRIDFACEEDGESRESPONSE._serialized_start=2025
  _GETGRIDFACEEDGESRESPONSE._serialized_end=2075
  _GETGRIDFACENODESRESPONSE._serialized_start=2077
  _GETGRIDFACENODESRESPONSE._serialized_end=2127
  _GETGRIDNODESPERFACERESPONSE._serialized_start=2129
  _GETGRIDNODESPERFACERESPONSE._serialized_end=2186
  _BMISERVICE._serialized_start=2189
  _BMISERVICE._serialized_end=4631
# @@protoc_insertion_point(module_scope)



from io import BytesIO

import docker
import numpy as np
import pytest
from typeguard import TypeCheckError

from grpc4bmi.bmi_client_docker import BmiClientDocker
from grpc4bmi.exceptions import DeadContainerException
from grpc4bmi.reserve import reserve_grid_padding

walrus_docker_image = 'ewatercycle/walrus-grpc4bmi:v0.3.1'


@pytest.fixture()
def walrus_model(tmp_path, walrus_input):
    model = BmiClientDocker(image=walrus_docker_image, image_port=55555, work_dir=str(tmp_path))
    yield model
    del model


@pytest.fixture()
def exit_container():
    client = docker.from_env()
    prog = "import sys;print('my stderr log message', file=sys.stderr);print('my stdout log message');sys.exit(25)"
    dockerfile = BytesIO(f'FROM {walrus_docker_image}\nCMD ["python3", "-c", "{prog}" ]\n'.encode())
    image, logs = client.images.build(fileobj=dockerfile)
    yield image.id
    client.images.remove(image.id, force=True)


@pytest.fixture()
def walrus_model_with_2input_dirs(tmp_path, walrus_2input_dirs):
    input_dirs = walrus_2input_dirs['input_dirs']
    work_dir = tmp_path / 'work'
    work_dir.mkdir()
    model = BmiClientDocker(image="ewatercycle/walrus-grpc4bmi:v0.2.0",
                            image_port=55555,
                            work_dir=str(work_dir),
                            input_dirs=input_dirs)
    yield model
    del model


class TestBmiClientDocker:
    def test_component_name(self, walrus_model):
        assert walrus_model.get_component_name() == 'WALRUS'

    def test_get_grid_origin(self, walrus_input, walrus_model):
        walrus_model.initialize(str(walrus_input))
        grid_id = walrus_model.get_var_grid('Q')

        result = reserve_grid_padding(walrus_model, grid_id)
        assert len(walrus_model.get_grid_origin(grid_id, result)) == 2

    def test_initialize(self, walrus_input, walrus_model):
        walrus_model.initialize(str(walrus_input))
        assert walrus_model.get_current_time() == walrus_model.get_start_time()

    def test_get_value_ptr(self, walrus_model):
        with pytest.raises(NotImplementedError):
            walrus_model.get_value_ptr('Q')

    def test_get_var_location(self, walrus_model):
        assert walrus_model.get_var_location('Q') == 'node'

    def test_2input_dirs(self, walrus_2input_dirs, walrus_model_with_2input_dirs):
        config_file = walrus_2input_dirs['cfg']
        walrus_model_with_2input_dirs.initialize(config_file)
        walrus_model_with_2input_dirs.update()
        # After initialization and update the forcings have been read from the extra volume
        val = walrus_model_with_2input_dirs.get_value('Q', np.zeros((1,)))
        assert len(val) == 1

    def test_inputdir_absent(self, tmp_path):
        dirthatdoesnotexist = 'dirthatdoesnotexist'
        input_dir = tmp_path / dirthatdoesnotexist
        work_dir = tmp_path / 'work'
        work_dir.mkdir()
        with pytest.raises(NotADirectoryError, match=dirthatdoesnotexist):
            BmiClientDocker(image=walrus_docker_image, image_port=55555,
                            work_dir=str(work_dir), input_dirs=[str(input_dir)])

    def test_workdir_absent(self, tmp_path):
        dirthatdoesnotexist = 'dirthatdoesnotexist'
        work_dir = tmp_path / dirthatdoesnotexist
        with pytest.raises(NotADirectoryError, match=dirthatdoesnotexist):
            BmiClientDocker(image=walrus_docker_image, image_port=55555, work_dir=str(work_dir))

    def test_container_start_failure(self, exit_container, tmp_path):
        expected = r"Failed to start Docker container with image"
        with pytest.raises(DeadContainerException, match=expected) as excinfo:
            BmiClientDocker(image=exit_container, work_dir=str(tmp_path))

        assert excinfo.value.exitcode == 25
        assert 'my stderr' in excinfo.value.logs
        assert 'my stdout' in excinfo.value.logs

    def test_same_inputdir_and_workdir(self, tmp_path):
        some_dir = str(tmp_path)
        match = 'Found work_dir equal to one of the input directories. Please drop that input dir.'
        with pytest.raises(ValueError, match=match):
            BmiClientDocker(image=walrus_docker_image, image_port=55555, input_dirs=(some_dir,), work_dir=some_dir)

    def test_workdir_as_number(self):
        with pytest.raises(TypeCheckError, match='is not an instance of str'):
            BmiClientDocker(image=walrus_docker_image, work_dir=42)

    def test_inputdirs_as_str(self, tmp_path):
        some_dir = str(tmp_path)
        with pytest.raises(TypeError, match='must be collections.abc.Iterable'):
            BmiClientDocker(image=walrus_docker_image, input_dirs='old type', work_dir=some_dir)

    def test_inputdirs_as_number(self, tmp_path):
        some_dir = str(tmp_path)
        with pytest.raises(TypeCheckError, match='is not an instance of collections.abc.Iterable'):
            BmiClientDocker(image=walrus_docker_image, input_dirs=42, work_dir=some_dir)

    def test_logs(self, walrus_model, capfd):
        logs = walrus_model.logs()

        assert 'R[write to console]' in logs
        assert 'R[write to console]' not in capfd.readouterr().out




class DeadContainerException(ChildProcessError):
    """
    Exception for when a container has died.

    Args:
        message (str): Human readable error message
        exitcode (int): The non-zero exit code of the container
        logs (str): Logs the container produced

    """
    def __init__(self, message, exitcode, logs, *args):
        super().__init__(message, *args)
        #: Exit code of container
        self.exitcode = exitcode
        #: Stdout and stderr of container
        self.logs = logs


class SingularityVersionException(ValueError):
    pass

class ApptainerVersionException(ValueError):
    pass



import numpy as np
import pytest

from grpc4bmi.run_server import BmiR, build_r

@pytest.fixture
def model():
    return build_r('FakeFailingRModel', 'test/fake.r')

@pytest.mark.skipif(not BmiR, reason='R and its dependencies are not installed')
class TestFakeFailingRModel:
    @pytest.mark.parametrize(
    'fn_name,fn_args',
    [
        ('get_component_name', tuple()),
        ('get_input_item_count', tuple()),
        ('get_output_item_count', tuple()),
        ('get_input_var_names', tuple()),
        ('get_output_var_names', tuple()),
        ('get_start_time', tuple()),
        ('get_end_time', tuple()),
        ('get_time_step', tuple()),
        ('get_time_units', tuple()),
        ('get_var_type', ['plate_surface__temperature']),
        ('get_var_units', ['plate_surface__temperature']),
        ('get_var_itemsize', ['plate_surface__temperature']),
        ('get_var_nbytes', ['plate_surface__temperature']),
        ('get_var_grid', ['plate_surface__temperature']),
        ('get_var_location', ['plate_surface__temperature']),
        ('get_grid_shape', [0, np.zeros((1,))]),
        ('get_grid_x', [0, np.zeros((1,))]),
        ('get_grid_y', [0, np.zeros((1,))]),
        ('get_grid_z', [0, np.zeros((1,))]),
        ('get_grid_spacing', [0, np.zeros((1,))]),
        ('get_grid_origin', [0, np.zeros((1,))]),
        ('get_grid_rank', [0]),
        ('get_grid_size', [0]),
        ('get_grid_type', [0]),
        ('initialize', ['config.ini']),
        ('update', tuple()),
        ('update_until', [2]),
        ('finalize', tuple()),
        ('get_current_time', tuple()),
        ('get_value', ['plate_surface__temperature', np.zeros((1,))]),
        ('get_value_at_indices', ['plate_surface__temperature', np.zeros((1,)), [1, 2, 3]]),
        ('set_value', ['plate_surface__temperature', np.ones((10, 20))]),
        ('set_value_at_indices', ['plate_surface__temperature', [1, 2, 3], [4, 5, 6]]),
        ('get_grid_node_count', (0,)),
        ('get_grid_edge_count', (0,)),
        ('get_grid_face_count', (0,)),
        ('get_grid_edge_nodes', (0, np.empty(0))),
        ('get_grid_face_nodes', (0, np.empty(0))),
        ('get_grid_face_edges', (0, np.empty(0))),
        ('get_grid_nodes_per_face', (42, np.empty(0))),
    ]
)
    def test_r_function_is_called(self, model: BmiR, fn_name, fn_args):
        # Every method in 'test/fake.r' executes the stop error action
        # So if no stop then no r function is called
        from rpy2.rinterface_lib.embedded import RRuntimeError
        with pytest.raises(RRuntimeError, match="Always fails"):
            fn = getattr(model, fn_name)
            fn(*fn_args)

    @pytest.mark.parametrize('fn_name,fn_args', [
        ('get_value_ptr', ['plate_surface__temperature']),
    ])
    def test_notimplemented(self, model: BmiR, fn_name, fn_args):
        with pytest.raises(NotImplementedError):
            fn = getattr(model, fn_name)
            fn(*fn_args)
        


import numpy as np
from heat import BmiHeat
from numpy.testing import assert_allclose
import pytest

from grpc4bmi.bmi_optionaldest import OptionalDestBmi
from test.fake_models import Rect3DGridModel, UnstructuredGridBmiModel


@pytest.mark.parametrize(
    'orig_model,method_name,method_args,expected_shape',
    [
        (BmiHeat(), 'get_value', ('plate_surface__temperature',), (200,)),
        (BmiHeat(), 'get_value', ('plate_surface__temperature', np.zeros(200)), (200,)),
        (BmiHeat(), 'get_grid_shape', (0,), (2,)),
        (BmiHeat(), 'get_grid_shape', (0, np.zeros(2)), (2,)),
        (BmiHeat(), 'get_grid_spacing', (0,), (2,)),
        (BmiHeat(), 'get_grid_spacing', (0, np.zeros(2)), (2,)),
        (BmiHeat(), 'get_grid_origin', (0,), (2,)),
        (BmiHeat(), 'get_grid_origin', (0, np.zeros(2)), (2,)),
        (BmiHeat(), 'get_value_at_indices', ['plate_surface__temperature', [1, 2, 3]], (3,)),
        (Rect3DGridModel(), 'get_grid_x', [0], (4,)),
        (Rect3DGridModel(), 'get_grid_x', [0, np.zeros((4,))], (4,)),
        (Rect3DGridModel(), 'get_grid_y', [0], (3,)),
        (Rect3DGridModel(), 'get_grid_y', [0, np.zeros((3,))], (3,)),
        (Rect3DGridModel(), 'get_grid_z', [0], (2,)),        
        (Rect3DGridModel(), 'get_grid_z', [0, np.zeros((2,))], (2,)),        
        (UnstructuredGridBmiModel(), 'get_grid_edge_nodes', [0], (16,)),
        (UnstructuredGridBmiModel(), 'get_grid_edge_nodes', [0, np.zeros((16,))], (16,)),
        (UnstructuredGridBmiModel(), 'get_grid_face_edges', [0], (11,)),
        (UnstructuredGridBmiModel(), 'get_grid_face_edges', [0, np.zeros((11,))], (11,)),
        (UnstructuredGridBmiModel(), 'get_grid_face_nodes', [0], (11,)),
        (UnstructuredGridBmiModel(), 'get_grid_face_nodes', [0, np.zeros((11,))], (11,)),
        (UnstructuredGridBmiModel(), 'get_grid_nodes_per_face', [0], (3,)),        
        (UnstructuredGridBmiModel(), 'get_grid_nodes_per_face', [0, np.zeros((3,))], (3,)),        
    ]
)
def test_methods_with_optional_dest(orig_model, method_name, method_args, expected_shape):
    model = OptionalDestBmi(orig_model)
    model.initialize(None)

    method = getattr(model, method_name)
    result = method(*method_args)

    assert result.shape == expected_shape


@pytest.mark.parametrize(
    'orig_model, method_name, method_args', 
    [
        (BmiHeat(), 'update', tuple()),
        (BmiHeat(), 'update_until', [2]),
        (BmiHeat(), 'finalize', tuple()),
        (BmiHeat(), 'get_current_time', tuple()),
        (BmiHeat(), 'get_component_name', tuple()),
        (BmiHeat(), 'get_input_item_count', tuple()),
        (BmiHeat(), 'get_output_item_count', tuple()),
        (BmiHeat(), 'get_input_var_names', tuple()),
        (BmiHeat(), 'get_output_var_names', tuple()),
        (BmiHeat(), 'get_start_time', tuple()),
        (BmiHeat(), 'get_end_time', tuple()),
        (BmiHeat(), 'get_time_step', tuple()),
        (BmiHeat(), 'get_time_units', tuple()),
        (BmiHeat(), 'get_var_type', ['plate_surface__temperature']),
        (BmiHeat(), 'get_var_units', ['plate_surface__temperature']),
        (BmiHeat(), 'get_var_itemsize', ['plate_surface__temperature']),
        (BmiHeat(), 'get_var_nbytes', ['plate_surface__temperature']),
        (BmiHeat(), 'get_var_grid', ['plate_surface__temperature']),
        (BmiHeat(), 'get_var_location', ['plate_surface__temperature']),
        (BmiHeat(), 'get_grid_rank', [0]),
        (BmiHeat(), 'get_grid_size', [0]),
        (BmiHeat(), 'get_grid_type', [0]),
        (BmiHeat(), 'get_value_ptr', ['plate_surface__temperature']),
        (BmiHeat(), 'set_value', ['plate_surface__temperature', np.ones((200,))]),
        (BmiHeat(), 'set_value_at_indices', ['plate_surface__temperature', [1, 2, 3], [4, 5, 6]]),
        (UnstructuredGridBmiModel(), 'get_grid_node_count', [0]),
        (UnstructuredGridBmiModel(), 'get_grid_edge_count', [0]),
        (UnstructuredGridBmiModel(), 'get_grid_face_count', [0]),
    ]
)
def test_methods_with_no_dest(orig_model, method_name, method_args):
    model = OptionalDestBmi(orig_model)
    model.initialize(None)

    method = getattr(model, method_name)
    result = method(*method_args)

    expected = getattr(orig_model, method_name)(*method_args)
    if isinstance(result, np.ndarray):
            assert_allclose(result, expected)
    else:
        assert result == expected



import logging
import math
import os
import socket
from contextlib import closing
from typing import Optional

import numpy as np
from bmipy import Bmi
import grpc
import numpy
from typeguard import typechecked

from grpc_status import rpc_status
from google.rpc import error_details_pb2

from . import bmi_pb2, bmi_pb2_grpc
from .constants import GRPC_MAX_MESSAGE_LENGTH

log = logging.getLogger(__name__)


class RemoteException(grpc.RpcError):
    def __init__(self, message, tb):
        super().__init__(message)
        self.remote_stacktrace = tb


def handle_error(exc):
    """Parsers DebugInfo (https://github.com/googleapis/googleapis/blob/07244bb797ddd6e0c1c15b02b4467a9a5729299f/google/rpc/error_details.proto#L46-L52) from the trailing metadata of a grpc.RpcError

    Args:
        exc (grpc.RpcError): Exception to handle

    Raises: original exception or RemoteException

    """
    status = rpc_status.from_call(exc)
    if status is None:
        raise
    for detail in status.details:
        if detail.Is(error_details_pb2.DebugInfo.DESCRIPTOR):
            info = error_details_pb2.DebugInfo()
            detail.Unpack(info)
            remote_traceback = info.stack_entries
            remote_detail = info.detail
            raise RemoteException(remote_detail, remote_traceback) from exc
    raise


def _fits_in_message(array):
    """Tests whether array can be passed through a gRPC message with a max message size of 4Mb"""
    array_size = array.size * array.itemsize
    return array_size <= GRPC_MAX_MESSAGE_LENGTH


class BmiClient(Bmi):
    """
    Client BMI interface, implementing BMI by forwarding every function call via GRPC to the server connected to the
    same port. A GRPC channel can be passed to the constructor; if not, it constructs an insecure channel on a free
    port itself. The timeout parameter indicates the model BMI startup timeout parameter (s).

    >>> import grpc
    >>> from grpc4bmi.bmi_grpc_client import BmiClient
    >>> mymodel = BmiClient(grpc.insecure_channel("localhost:<PORT>"))
    >>> print(mymodel.get_component_name())
    Hello world
    """

    def __init__(self, channel=None, timeout=None, stub=None):
        if stub is None:
            c = BmiClient.create_grpc_channel() if channel is None else channel
            self.stub = bmi_pb2_grpc.BmiServiceStub(c)
            future = grpc.channel_ready_future(c)
            future.result(timeout=timeout)
        else:
            self.stub = stub

    def __del__(self):
        del self.stub

    @staticmethod
    def create_grpc_channel(port=0, host=None):
        p, h = port, host
        if h is None:
            h = "localhost"
        if p == 0:
            p = os.environ.get("BMI_PORT", 50051)
        return grpc.insecure_channel(':'.join([h, str(p)]))

    @staticmethod
    def get_unique_port(host=None):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("" if host is None else host, 0))
            return int(s.getsockname()[1])

    @typechecked
    def initialize(self, filename: Optional[str]):
        fname = "" if filename is None else filename
        try:
            return self.stub.initialize(bmi_pb2.InitializeRequest(config_file=fname))
        except grpc.RpcError as e:
            handle_error(e)

    def update(self):
        try:
            self.stub.update(bmi_pb2.Empty())
        except grpc.RpcError as e:
            handle_error(e)

    def update_until(self, time: float) -> None:
        try:
            self.stub.updateUntil(bmi_pb2.GetTimeResponse(time=time))
        except grpc.RpcError as e:
            handle_error(e)

    def finalize(self):
        try:
            self.stub.finalize(bmi_pb2.Empty())
        except grpc.RpcError as e:
            handle_error(e)

    def get_component_name(self):
        try:
            return str(self.stub.getComponentName(bmi_pb2.Empty()).name)
        except grpc.RpcError as e:
            handle_error(e)

    def get_input_item_count(self) -> int:
        try:
            return self.stub.getInputItemCount(bmi_pb2.Empty()).count
        except grpc.RpcError as e:
            handle_error(e)

    def get_output_item_count(self) -> int:
        try:
            return self.stub.getOutputItemCount(bmi_pb2.Empty()).count
        except grpc.RpcError as e:
            handle_error(e)

    def get_input_var_names(self):
        try:
            return tuple([str(s) for s in self.stub.getInputVarNames(bmi_pb2.Empty()).names])
        except grpc.RpcError as e:
            handle_error(e)

    def get_output_var_names(self):
        try:
            return tuple([str(s) for s in self.stub.getOutputVarNames(bmi_pb2.Empty()).names])
        except grpc.RpcError as e:
            handle_error(e)

    def get_time_units(self):
        try:
            response = str(self.stub.getTimeUnits(bmi_pb2.Empty()).units)
            return None if not response else response
        except grpc.RpcError as e:
            handle_error(e)

    def get_time_step(self):
        try:
            return self.stub.getTimeStep(bmi_pb2.Empty()).interval
        except grpc.RpcError as e:
            handle_error(e)

    def get_current_time(self):
        try:
            return self.stub.getCurrentTime(bmi_pb2.Empty()).time
        except grpc.RpcError as e:
            handle_error(e)

    def get_start_time(self):
        try:
            return self.stub.getStartTime(bmi_pb2.Empty()).time
        except grpc.RpcError as e:
            handle_error(e)

    def get_end_time(self):
        try:
            return self.stub.getEndTime(bmi_pb2.Empty()).time
        except grpc.RpcError as e:
            handle_error(e)

    def get_var_grid(self, name):
        try:
            return self.stub.getVarGrid(bmi_pb2.GetVarRequest(name=name)).grid_id
        except grpc.RpcError as e:
            handle_error(e)

    def get_var_type(self, name):
        try:
            return str(self.stub.getVarType(bmi_pb2.GetVarRequest(name=name)).type)
        except grpc.RpcError as e:
            handle_error(e)

    def get_var_itemsize(self, name):
        try:
            item_size = self.stub.getVarItemSize(bmi_pb2.GetVarRequest(name=name)).size
            if item_size == 0:
                # BMI < v2.0 did not have get_var_itemsize, so old server will return 0
                # fallback to getting item size from var type
                var_type = self.get_var_type(name)
                try:
                    item_size = numpy.dtype(var_type).itemsize
                    log.info(f'get_var_itemsize returned 0, corrected to {item_size} using get_var_type.')
                except TypeError:
                    raise ValueError('get_var_itemsize returned 0, which is impossible')
            return item_size
        except grpc.RpcError as e:
            handle_error(e)

    def get_var_units(self, name):
        try:
            response = str(self.stub.getVarUnits(bmi_pb2.GetVarRequest(name=name)).units)
            return None if not response else response
        except grpc.RpcError as e:
            handle_error(e)

    def get_var_nbytes(self, name):
        try:
            return self.stub.getVarNBytes(bmi_pb2.GetVarRequest(name=name)).nbytes
        except grpc.RpcError as e:
            handle_error(e)

    def get_var_location(self, name: str) -> str:
        try:
            location = self.stub.getVarLocation(bmi_pb2.GetVarRequest(name=name)).location
            return bmi_pb2.GetVarLocationResponse.Location.Name(location).lower()
        except grpc.RpcError as e:
            handle_error(e)

    def get_value(self, name, dest):
        fits = _fits_in_message(dest)
        if not fits:
            return self._chunked_get_value(name, dest)
        try:
            response = self.stub.getValue(bmi_pb2.GetVarRequest(name=name))
            numpy.copyto(src=BmiClient.make_array(response), dst=dest)
            return dest
        except grpc.RpcError as e:
            handle_error(e)

    def _chunked_get_value(self, name: str, dest: np.array) -> np.array:
        # Make chunk one item smaller than maximum (4Mb)
        chunk_size = math.floor(GRPC_MAX_MESSAGE_LENGTH / dest.dtype.itemsize) - dest.dtype.itemsize
        chunks = []
        log.info(f'Too many items ({dest.size}) for single call, '
                 f'using multiple get_value_at_indices() with into chunks of {chunk_size} items')
        for i in range(0, dest.size, chunk_size):
            start = i
            stop = i + chunk_size
            # Last chunk can be smaller
            if stop > dest.size:
                stop = dest.size
            chunks.append(self._get_value_at_range(name, start, stop))

        numpy.concatenate(chunks, out=dest)
        return dest

    def _get_value_at_range(self, name, start, stop):
        log.info(f'Fetching value range {start} - {stop}')
        try:
            response = self.stub.getValueAtIndices(bmi_pb2.GetValueAtIndicesRequest(name=name,
                                                                                    indices=range(start, stop)))
            return BmiClient.make_array(response)
        except grpc.RpcError as e:
            handle_error(e)

    def get_value_ptr(self, name: str) -> np.ndarray:
        """Not possible, unable give reference to data structure in another process and possibly another machine"""
        raise NotImplementedError("Array references cannot be transmitted through this GRPC channel")

    def get_value_at_indices(self, name, dest, indices):
        try:
            index_array = indices
            if indices is list:
                index_array = numpy.array(indices)
            response = self.stub.getValueAtIndices(bmi_pb2.GetValueAtIndicesRequest(name=name,
                                                                                    indices=index_array.flatten()))
            numpy.copyto(src=BmiClient.make_array(response), dst=dest)
            return dest
        except grpc.RpcError as e:
            handle_error(e)

    def set_value(self, name, values):
        try:
            if values.dtype in (numpy.int16, numpy.int32, numpy.int64):
                request = bmi_pb2.SetValueRequest(name=name,
                                                  values_int=bmi_pb2.IntArrayMessage(values=values.flatten()))
            elif values.dtype in (numpy.float32, numpy.float16):
                request = bmi_pb2.SetValueRequest(name=name,
                                                  values_float=bmi_pb2.FloatArrayMessage(values=values.flatten()))
            elif values.dtype == numpy.float64:
                request = bmi_pb2.SetValueRequest(name=name,
                                                  values_double=bmi_pb2.DoubleArrayMessage(values=values.flatten()))
            else:
                raise NotImplementedError("Arrays with type %s cannot be transmitted through this GRPC channel" % values.dtype)
            self.stub.setValue(request)
        except grpc.RpcError as e:
            handle_error(e)

    def set_value_at_indices(self, name, inds, src):
        try:
            index_array = inds
            if inds is list:
                index_array = numpy.array(inds)
            if src.dtype in (numpy.int32, numpy.int64):
                request = bmi_pb2.SetValueAtIndicesRequest(name=name,
                                                           indices=index_array.flatten(),
                                                           values_int=bmi_pb2.IntArrayMessage(values=src.flatten()))
            elif src.dtype in (numpy.float32, numpy.float16):
                request = bmi_pb2.SetValueAtIndicesRequest(name=name,
                                                           indices=index_array.flatten(),
                                                           values_float=bmi_pb2.FloatArrayMessage(values=src.flatten()))
            elif src.dtype == numpy.float64:
                request = bmi_pb2.SetValueAtIndicesRequest(name=name,
                                                           indices=index_array.flatten(),
                                                           values_double=bmi_pb2.DoubleArrayMessage(values=src.flatten()))
            else:
                raise NotImplementedError("Arrays with type %s cannot be transmitted through this GRPC channel" % src.dtype)
            self.stub.setValueAtIndices(request)
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_size(self, grid):
        try:
            return self.stub.getGridSize(bmi_pb2.GridRequest(grid_id=grid)).size
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_rank(self, grid):
        try:
            return self.stub.getGridRank(bmi_pb2.GridRequest(grid_id=grid)).rank
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_type(self, grid):
        try:
            return str(self.stub.getGridType(bmi_pb2.GridRequest(grid_id=grid)).type)
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_x(self, grid, x):
        try:
            src = numpy.array(self.stub.getGridX(bmi_pb2.GridRequest(grid_id=grid)).coordinates)
            numpy.copyto(src=src, dst=x)
            return x
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_y(self, grid, y):
        try:
            src = numpy.array(self.stub.getGridY(bmi_pb2.GridRequest(grid_id=grid)).coordinates)
            numpy.copyto(src=src, dst=y)
            return y
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_z(self, grid, z):
        try:
            src = numpy.array(self.stub.getGridZ(bmi_pb2.GridRequest(grid_id=grid)).coordinates)
            numpy.copyto(src=src, dst=z)
            return z
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_shape(self, grid, shape):
        try:
            src = tuple(self.stub.getGridShape(bmi_pb2.GridRequest(grid_id=grid)).shape)
            numpy.copyto(src=src, dst=shape)
            return shape
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_spacing(self, grid, spacing):
        try:
            src = tuple(self.stub.getGridSpacing(bmi_pb2.GridRequest(grid_id=grid)).spacing)
            numpy.copyto(src=src, dst=spacing)
            return spacing
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_origin(self, grid, origin):
        try:
            src = tuple(self.stub.getGridOrigin(bmi_pb2.GridRequest(grid_id=grid)).origin)
            numpy.copyto(src=src, dst=origin)
            return origin
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_node_count(self, grid: int) -> int:
        try:
            return self.stub.getGridNodeCount(bmi_pb2.GridRequest(grid_id=grid)).count
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_edge_count(self, grid: int) -> int:
        try:
            return self.stub.getGridEdgeCount(bmi_pb2.GridRequest(grid_id=grid)).count
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_face_count(self, grid: int) -> int:
        try:
            return self.stub.getGridFaceCount(bmi_pb2.GridRequest(grid_id=grid)).count
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_edge_nodes(self, grid: int, edge_nodes: np.ndarray) -> np.ndarray:
        try:
            links = self.stub.getGridEdgeNodes(bmi_pb2.GridRequest(grid_id=grid)).edge_nodes
            numpy.copyto(src=links, dst=edge_nodes)
            return edge_nodes
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_face_nodes(self, grid: int, face_nodes: np.ndarray) -> np.ndarray:
        try:
            links = self.stub.getGridFaceNodes(bmi_pb2.GridRequest(grid_id=grid)).face_nodes
            numpy.copyto(src=links, dst=face_nodes)
            return face_nodes
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_face_edges(self, grid: int, face_edges: np.ndarray) -> np.ndarray:
        try:
            links = self.stub.getGridFaceEdges(bmi_pb2.GridRequest(grid_id=grid)).face_edges
            numpy.copyto(src=links, dst=face_edges)
            return face_edges
        except grpc.RpcError as e:
            handle_error(e)

    def get_grid_nodes_per_face(self, grid: int, nodes_per_face: np.ndarray) -> np.ndarray:
        try:
            links = self.stub.getGridNodesPerFace(bmi_pb2.GridRequest(grid_id=grid)).nodes_per_face
            numpy.copyto(src=links, dst=nodes_per_face)
            return nodes_per_face
        except grpc.RpcError as e:
            handle_error(e)

    @staticmethod
    def make_array(response):
        if response.HasField("values_int"):
            return numpy.array(response.values_int.values)
        if response.HasField("values_float"):
            return numpy.array(response.values_float.values)
        if response.HasField("values_double"):
            return numpy.array(response.values_double.values)



from bmipy import Bmi


class MemoizedBmi(Bmi):
    """Wrapper around Bmi object that caches the return values of almost all methods.

    Most BMI methods return same value each time it is called, so the results can be cached.
    gRPC communication is expensive and can be sped up by caching.

    The following methods are not cached:

    * initialize
    * update_*
    * finalize
    * get_current_time
    * get_value_*
    * set_value_*

    The cache is cleared when initialize() is called.

    Example:

        A gRPC BMI server is running on localhost:1234, to cache it use the following.

        >>> import grpc
        >>> from grpc4bmi.bmi_grpc_client import BmiClient
        >>> from grpc4bmi.bmi_memoized import MemoizedBmi
        >>> slow_model = BmiClient(grpc.insecure_channel("localhost:1234"))
        >>> model = MemoizedBmi(slow_model)
        >>> print(model.get_component_name())
        Hello world
        >>> # Calling second time will return cached value
        >>> # and not talk to server on "localhost:1234"
        >>> print(model.get_component_name())
        Hello world

    """
    def __init__(self, origin: Bmi):
        self.origin = origin
        self.cache = dict()

    def _cache(self, fn, arg=None, output=None):
        if fn not in self.cache:
            self.cache[fn] = dict()
        if arg not in self.cache[fn]:
            if arg is None:
                self.cache[fn][arg] = getattr(self.origin, fn)()
            else:
                if output is None:
                    self.cache[fn][arg] = getattr(self.origin, fn)(arg)
                else:
                    self.cache[fn][arg] = getattr(self.origin, fn)(arg, output)
        return self.cache[fn][arg]

    def initialize(self, filename):
        self.cache = dict()
        return self.origin.initialize(filename)

    def update(self):
        self.origin.update()

    def update_until(self, time):
        self.origin.update_until(time)

    def finalize(self):
        self.origin.finalize()

    def get_component_name(self):
        return self._cache('get_component_name')

    def get_input_item_count(self):
        return self._cache('get_input_item_count')

    def get_output_item_count(self):
        return self._cache('get_output_item_count')

    def get_input_var_names(self):
        return self._cache('get_input_var_names')

    def get_output_var_names(self):
        return self._cache('get_output_var_names')

    def get_start_time(self):
        return self._cache('get_start_time')

    def get_current_time(self):
        return self.origin.get_current_time()

    def get_end_time(self):
        return self._cache('get_end_time')

    def get_time_step(self):
        return self._cache('get_time_step')

    def get_time_units(self):
        return self._cache('get_time_units')

    def get_var_type(self, var_name):
        return self._cache('get_var_type', var_name)

    def get_var_units(self, var_name):
        return self._cache('get_var_units', var_name)

    def get_var_itemsize(self, var_name):
        return self._cache('get_var_itemsize', var_name)

    def get_var_nbytes(self, var_name):
        return self._cache('get_var_nbytes', var_name)

    def get_var_location(self, var_name):
        return self._cache('get_var_location', var_name)

    def get_var_grid(self, var_name):
        return self._cache('get_var_grid', var_name)

    def get_value(self, var_name, dest):
        return self.origin.get_value(var_name, dest)

    def get_value_ptr(self, var_name):
        return self.origin.get_value_ptr(var_name)  

    def get_value_at_indices(self, var_name, dest, inds):
        return self.origin.get_value_at_indices(var_name, dest, inds)

    def set_value(self, var_name, src):
        return self.origin.set_value(var_name, src)

    def set_value_at_indices(self, var_name, indices, src):
        return self.origin.set_value_at_indices(var_name, indices, src)

    def get_grid_shape(self, grid, shape):
        shape[:] = self._cache('get_grid_shape', grid, shape)
        return shape

    def get_grid_x(self, grid, x):
        x[:] = self._cache('get_grid_x', grid, x)
        return x

    def get_grid_y(self, grid, y):
        y[:] = self._cache('get_grid_y', grid, y)
        return y

    def get_grid_z(self, grid, z):
        z[:] = self._cache('get_grid_z', grid, z)
        return z

    def get_grid_spacing(self, grid, spacing):
        spacing[:] = self._cache('get_grid_spacing', grid, spacing)
        return spacing

    def get_grid_origin(self, grid, origin):
        origin[:] = self._cache('get_grid_origin', grid, origin)
        return origin

    def get_grid_rank(self, grid):
        return self._cache('get_grid_rank', grid)

    def get_grid_size(self, grid):
        return self._cache('get_grid_size', grid)

    def get_grid_type(self, grid):
        return self._cache('get_grid_type', grid)

    def get_grid_node_count(self, grid):
        return self._cache('get_grid_node_count', grid)

    def get_grid_edge_count(self, grid):
        return self._cache('get_grid_edge_count', grid)

    def get_grid_face_count(self, grid):
        return self._cache('get_grid_face_count', grid)

    def get_grid_edge_nodes(self, grid, edge_nodes):
        edge_nodes[:] = self._cache('get_grid_edge_nodes', grid, edge_nodes)
        return edge_nodes

    def get_grid_face_edges(self, grid, face_edges):
        face_edges[:] = self._cache('get_grid_face_edges', grid, face_edges)
        return face_edges

    def get_grid_face_nodes(self, grid, face_nodes):
        face_nodes[:] = self._cache('get_grid_face_nodes', grid, face_nodes)
        return face_nodes
    
    def get_grid_nodes_per_face(self, grid, nodes_per_face):
        nodes_per_face[:] = self._cache('get_grid_nodes_per_face', grid, nodes_per_face)
        return nodes_per_face



import os
import time
from os.path import abspath
from typing import Iterable

import docker
from docker.models.containers import Container
from typeguard import typechecked

from grpc4bmi.bmi_grpc_client import BmiClient
from grpc4bmi.exceptions import DeadContainerException


class BmiClientDocker(BmiClient):
    """
    BMI gRPC client for dockerized server processes: the initialization launches the docker container which should have the
    run-bmi-server as its command. Also, it should expose the tcp port 50001 for communication with this client. Upon
    destruction, this class terminates the corresponding docker server.

    Args:
        image (str): Docker image name of grpc4bmi wrapped model
        image_port (int): Port of server inside the image
        host (str): Host on which the image port is published on a random port
        input_dirs (Iterable[str]): Directories for input files of model.

            All of directories will be mounted read-only inside Docker container on same path as outside container.

        work_dir (str): Working directory for model.

            Directory is mounted inside container and changed into.

        user (str): Username or UID of Docker container. Defaults to own UID.
        remove (bool): Automatically remove the container and logs when it exits.

            Enable to get logs when container dies prematurely.

        delay (int): Seconds to wait for Docker container to startup, before connecting to it
        timeout (int): Seconds to wait for gRPC client to connect to server
        extra_volumes (Dict[str,Dict]): Extra volumes to attach to Docker container.
            The key is either the hosts path or a volume name and the value is a dictionary with the keys:

            - ``bind`` The path to mount the volume inside the container
            - ``mode`` Either ``rw`` to mount the volume read/write, or ``ro`` to mount it read-only.

            For example:

            .. code-block:: python

                {'/data/shared/forcings/': {'bind': '/forcings', 'mode': 'ro'}}

            Increase when container takes a long time to startup.

        timeout (int): Seconds to wait for gRPC client to connect to server.

            By default will try forever to connect to gRPC server inside container.
            Set to low number to escape endless wait.

    See :py:class:`grpc4bmi.bmi_client_apptainer.BmiClientApptainer` for examples using `input_dirs` and `work_dir`.
    """

    @typechecked
    def __init__(self, image: str, work_dir: str, image_port=50051, host=None,
                 input_dirs: Iterable[str] = tuple(),
                 user=os.getuid(), remove=False, delay=5,
                 timeout=None):
        if type(input_dirs) == str:
            msg = f'type of argument "input_dirs" must be collections.abc.Iterable; ' \
                  f'got {type(input_dirs)} instead'
            raise TypeError(msg)
        port = BmiClient.get_unique_port()
        client = docker.from_env()
        volumes = {}
        for raw_input_dir in input_dirs:
            input_dir = abspath(raw_input_dir)
            if not os.path.isdir(input_dir):
                raise NotADirectoryError(input_dir)
            volumes[input_dir] = {"bind": input_dir, "mode": "ro"}

        self.work_dir = abspath(work_dir)
        if self.work_dir in volumes:
            raise ValueError('Found work_dir equal to one of the input directories. Please drop that input dir.')
        if not os.path.isdir(self.work_dir):
            raise NotADirectoryError(self.work_dir)
        volumes[self.work_dir] = {"bind": self.work_dir, "mode": "rw"}
        self.container: Container = client.containers.run(image,
                                               ports={str(image_port) + "/tcp": port},
                                               volumes=volumes,
                                               working_dir=self.work_dir,
                                               user=user,
                                               remove=remove,
                                               detach=True)
        time.sleep(delay)
        if not remove:
            # Only able to reload, read logs on exited container when remove=False
            self.container.reload()
            if self.container.status == 'exited':
                exitcode = self.container.attrs["State"]["ExitCode"]
                logs = self.logs()
                msg = f'Failed to start Docker container with image {image}, Container log: {logs}'
                raise DeadContainerException(msg, exitcode, logs)

        super(BmiClientDocker, self).__init__(BmiClient.create_grpc_channel(port=port, host=host), timeout=timeout)

    def __del__(self):
        self.container.stop()

    def logs(self) -> str:
        """Returns complete combined stdout and stderr written by the Docker container.
        """
        return self.container.logs().decode('utf8')



from typing import Optional, Tuple
from bmipy import Bmi
import numpy as np

from grpc4bmi.reserve import reserve_grid_edge_nodes, reserve_grid_face_, reserve_grid_nodes, reserve_grid_padding, reserve_grid_shape, reserve_grid_nodes_per_face, reserve_values, reserve_values_at_indices

class OptionalDestBmi(Bmi):
    """Proxy around Bmi object that makes the output argument (dest) optional.

    Some Bmi methods accept a numpy array as argument (dest) that must be filled by the model.
    This class makes that argument optional.

    For example to get the value of a `heat model <https://github.com/csdms/bmi-example-python>`_:

    .. code-block:: python

            from heat import BmiHeat
            import numpy as np

            orig_model = BmiHeat()
            orig_model.initialize()
            dest = np.empty((200,), dtype=np.float64)
            dest2 = orig_model.get_value('plate_surface__temperature', dest)
            

    A `dest` variable must created and passed to the `get_value` method.
    It must be set to a Numpy array with the shape and type the model requires.
    The returned numpy array (`dest2`) is the same as the `dest` numpy array.
    By using this class we no longer need do create and pass it.

    .. code-block:: python

        from grpc4bmi.bmi_optionaldest import OptionalDestBmi

        model = OptionalDestBmi(orig_model)
        dest3 = model.get_value('plate_surface__temperature')

    This class will create dest variables by calling other Bmi methods for its shape.
    When using a slow model it is strongly suggested to wrap the BmiClient with :class:`grpc4bmi.bmi_memoized.MemoizedBmi` first to prevent a lot of calls.

    The `dest` in :external+bmipy:ref:`bmipy.Bmi.get_value_at_indices(self, name, dest, inds) <get_value_at_indices>` method
    can not be made optional so it has been changed to
    :func:`OptionalDestBmi.get_value_at_indices(self, name, inds) <grpc4bmi.bmi_optionaldest.OptionalDestBmi.get_value_at_indices>`.
    """
    def __init__(self, origin: Bmi):
        self.origin = origin

    def initialize(self, config_file: Optional[str]) -> None:
        return self.origin.initialize(config_file)

    def update(self) -> None:
        self.origin.update()

    def update_until(self, time) -> None:
        self.origin.update_until(time)

    def finalize(self) -> None:
        self.origin.finalize()

    def get_component_name(self) -> str:
        return self.origin.get_component_name()

    def get_input_item_count(self) -> int:
        return self.origin.get_input_item_count()

    def get_output_item_count(self) -> int:
        return self.origin.get_output_item_count()

    def get_input_var_names(self) -> Tuple[str]:
        return self.origin.get_input_var_names()

    def get_output_var_names(self) -> Tuple[str]:
        return self.origin.get_output_var_names()

    def get_start_time(self) -> float:
        return self.origin.get_start_time()

    def get_current_time(self) -> float:
        return self.origin.get_current_time()

    def get_end_time(self) -> float:
        return self.origin.get_end_time()

    def get_time_step(self) -> float:
        return self.origin.get_time_step()

    def get_time_units(self) -> str:
        return self.origin.get_time_units()

    def get_var_type(self, name: str) -> str:
        return self.origin.get_var_type(name)

    def get_var_units(self, name: str) -> str:
        return self.origin.get_var_units(name)

    def get_var_itemsize(self, name: str) -> int:
        return self.origin.get_var_itemsize(name)

    def get_var_nbytes(self, name: str) -> int:
        return self.origin.get_var_nbytes(name)

    def get_var_location(self, name: str) -> str:
        return self.origin.get_var_location(name)

    def get_var_grid(self, name: str) -> int:
        return self.origin.get_var_grid(name)

    def get_value(self, name, dest: Optional[np.ndarray]=None) -> np.ndarray:
        if dest is None:
            dest = reserve_values(self.origin, name)
        return self.origin.get_value(name, dest)

    def get_value_ptr(self, name: str) -> np.ndarray:
        return self.origin.get_value_ptr(name)  

    def get_value_at_indices(self, name: str, inds: np.ndarray) -> np.ndarray:
        """Get values at particular indices.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.
        inds : array_like
            The indices into the variable array.

        Returns
        -------
        array_like
            Value of the model variable at the given location.
        """
        dest = reserve_values_at_indices(self.origin, name, inds)
        return self.origin.get_value_at_indices(name, dest, inds)

    def set_value(self, name: str, values: np.ndarray) -> None:
        return self.origin.set_value(name, values)

    def set_value_at_indices(self, name: str, inds: np.ndarray, src: np.ndarray) -> None:
        return self.origin.set_value_at_indices(name, inds, src)

    def get_grid_shape(self, grid: int, shape: Optional[np.ndarray]=None) -> np.ndarray:
        if shape is None:
            shape = reserve_grid_shape(self.origin, grid)
        return self.origin.get_grid_shape(grid, shape)

    def get_grid_x(self, grid: int, x: Optional[np.ndarray]=None) -> np.ndarray:
        if x is None:
            x = reserve_grid_nodes(self.origin, grid, 0)
        return self.origin.get_grid_x(grid, x)

    def get_grid_y(self, grid: int, y: Optional[np.ndarray]=None) -> np.ndarray:
        if y is None:
            y = reserve_grid_nodes(self.origin, grid, 1)
        return self.origin.get_grid_y(grid, y)

    def get_grid_z(self, grid: int, z: Optional[np.ndarray]=None) -> np.ndarray:
        if z is None:
            z = reserve_grid_nodes(self.origin, grid, 2)
        return self.origin.get_grid_z(grid, z)

    def get_grid_spacing(self, grid: int, spacing: Optional[np.ndarray]=None) -> np.ndarray:
        if spacing is None:
            spacing = reserve_grid_padding(self.origin, grid)
        return self.origin.get_grid_spacing(grid, spacing)

    def get_grid_origin(self, grid: int, origin: Optional[np.ndarray]=None) -> np.ndarray:
        if origin is None:
            origin = reserve_grid_padding(self.origin, grid)
        return self.origin.get_grid_origin(grid, origin)

    def get_grid_rank(self, grid: int) -> int:
        return self.origin.get_grid_rank(grid)

    def get_grid_size(self, grid: int) -> int:
        return self.origin.get_grid_size(grid)

    def get_grid_type(self, grid: int) -> str:
        return self.origin.get_grid_type(grid)

    def get_grid_node_count(self, grid: int) -> int:
        return self.origin.get_grid_node_count(grid)

    def get_grid_edge_count(self, grid: int) -> int:
        return self.origin.get_grid_edge_count(grid)

    def get_grid_face_count(self, grid: int) -> int:
        return self.origin.get_grid_face_count(grid)

    def get_grid_edge_nodes(self, grid: int, edge_nodes: Optional[np.ndarray]=None) -> np.ndarray:
        if edge_nodes is None:
            edge_nodes = reserve_grid_edge_nodes(self.origin, grid)
        return self.origin.get_grid_edge_nodes(grid, edge_nodes)

    def get_grid_face_edges(self, grid: int, face_edges: Optional[np.ndarray]=None) -> np.ndarray:
        if face_edges is None:
            face_edges = reserve_grid_face_(self.origin, grid)
        return self.origin.get_grid_face_edges(grid, face_edges)

    def get_grid_face_nodes(self, grid: int, face_nodes: Optional[np.ndarray]=None) -> np.ndarray:
        if face_nodes is None:
            face_nodes = reserve_grid_face_(self.origin, grid)
        return self.origin.get_grid_face_nodes(grid, face_nodes)
    
    def get_grid_nodes_per_face(self, grid: int, nodes_per_face: Optional[np.ndarray]=None) -> np.ndarray:
        if nodes_per_face is None:
            nodes_per_face = reserve_grid_nodes_per_face(self.origin, grid)
        return self.origin.get_grid_nodes_per_face(grid, nodes_per_face)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from grpc4bmi import bmi_pb2 as grpc4bmi_dot_bmi__pb2


class BmiServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.initialize = channel.unary_unary(
                '/bmi.BmiService/initialize',
                request_serializer=grpc4bmi_dot_bmi__pb2.InitializeRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                )
        self.update = channel.unary_unary(
                '/bmi.BmiService/update',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                )
        self.updateUntil = channel.unary_unary(
                '/bmi.BmiService/updateUntil',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                )
        self.finalize = channel.unary_unary(
                '/bmi.BmiService/finalize',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                )
        self.getComponentName = channel.unary_unary(
                '/bmi.BmiService/getComponentName',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetComponentNameResponse.FromString,
                )
        self.getInputItemCount = channel.unary_unary(
                '/bmi.BmiService/getInputItemCount',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
                )
        self.getOutputItemCount = channel.unary_unary(
                '/bmi.BmiService/getOutputItemCount',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
                )
        self.getInputVarNames = channel.unary_unary(
                '/bmi.BmiService/getInputVarNames',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarNamesResponse.FromString,
                )
        self.getOutputVarNames = channel.unary_unary(
                '/bmi.BmiService/getOutputVarNames',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarNamesResponse.FromString,
                )
        self.getTimeUnits = channel.unary_unary(
                '/bmi.BmiService/getTimeUnits',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetTimeUnitsResponse.FromString,
                )
        self.getTimeStep = channel.unary_unary(
                '/bmi.BmiService/getTimeStep',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetTimeStepResponse.FromString,
                )
        self.getCurrentTime = channel.unary_unary(
                '/bmi.BmiService/getCurrentTime',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
                )
        self.getStartTime = channel.unary_unary(
                '/bmi.BmiService/getStartTime',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
                )
        self.getEndTime = channel.unary_unary(
                '/bmi.BmiService/getEndTime',
                request_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
                )
        self.getVarGrid = channel.unary_unary(
                '/bmi.BmiService/getVarGrid',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarGridResponse.FromString,
                )
        self.getVarType = channel.unary_unary(
                '/bmi.BmiService/getVarType',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarTypeResponse.FromString,
                )
        self.getVarItemSize = channel.unary_unary(
                '/bmi.BmiService/getVarItemSize',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarItemSizeResponse.FromString,
                )
        self.getVarUnits = channel.unary_unary(
                '/bmi.BmiService/getVarUnits',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarUnitsResponse.FromString,
                )
        self.getVarNBytes = channel.unary_unary(
                '/bmi.BmiService/getVarNBytes',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarNBytesResponse.FromString,
                )
        self.getVarLocation = channel.unary_unary(
                '/bmi.BmiService/getVarLocation',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetVarLocationResponse.FromString,
                )
        self.getValue = channel.unary_unary(
                '/bmi.BmiService/getValue',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetValueResponse.FromString,
                )
        self.getValueAtIndices = channel.unary_unary(
                '/bmi.BmiService/getValueAtIndices',
                request_serializer=grpc4bmi_dot_bmi__pb2.GetValueAtIndicesRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetValueAtIndicesResponse.FromString,
                )
        self.setValue = channel.unary_unary(
                '/bmi.BmiService/setValue',
                request_serializer=grpc4bmi_dot_bmi__pb2.SetValueRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                )
        self.setValueAtIndices = channel.unary_unary(
                '/bmi.BmiService/setValueAtIndices',
                request_serializer=grpc4bmi_dot_bmi__pb2.SetValueAtIndicesRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                )
        self.getGridSize = channel.unary_unary(
                '/bmi.BmiService/getGridSize',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridSizeResponse.FromString,
                )
        self.getGridType = channel.unary_unary(
                '/bmi.BmiService/getGridType',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridTypeResponse.FromString,
                )
        self.getGridRank = channel.unary_unary(
                '/bmi.BmiService/getGridRank',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridRankResponse.FromString,
                )
        self.getGridShape = channel.unary_unary(
                '/bmi.BmiService/getGridShape',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridShapeResponse.FromString,
                )
        self.getGridSpacing = channel.unary_unary(
                '/bmi.BmiService/getGridSpacing',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridSpacingResponse.FromString,
                )
        self.getGridOrigin = channel.unary_unary(
                '/bmi.BmiService/getGridOrigin',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridOriginResponse.FromString,
                )
        self.getGridX = channel.unary_unary(
                '/bmi.BmiService/getGridX',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.FromString,
                )
        self.getGridY = channel.unary_unary(
                '/bmi.BmiService/getGridY',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.FromString,
                )
        self.getGridZ = channel.unary_unary(
                '/bmi.BmiService/getGridZ',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.FromString,
                )
        self.getGridNodeCount = channel.unary_unary(
                '/bmi.BmiService/getGridNodeCount',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
                )
        self.getGridEdgeCount = channel.unary_unary(
                '/bmi.BmiService/getGridEdgeCount',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
                )
        self.getGridFaceCount = channel.unary_unary(
                '/bmi.BmiService/getGridFaceCount',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
                )
        self.getGridEdgeNodes = channel.unary_unary(
                '/bmi.BmiService/getGridEdgeNodes',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridEdgeNodesResponse.FromString,
                )
        self.getGridFaceNodes = channel.unary_unary(
                '/bmi.BmiService/getGridFaceNodes',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridFaceNodesResponse.FromString,
                )
        self.getGridFaceEdges = channel.unary_unary(
                '/bmi.BmiService/getGridFaceEdges',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridFaceEdgesResponse.FromString,
                )
        self.getGridNodesPerFace = channel.unary_unary(
                '/bmi.BmiService/getGridNodesPerFace',
                request_serializer=grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
                response_deserializer=grpc4bmi_dot_bmi__pb2.GetGridNodesPerFaceResponse.FromString,
                )


class BmiServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def initialize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def update(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def updateUntil(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def finalize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getComponentName(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getInputItemCount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getOutputItemCount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getInputVarNames(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getOutputVarNames(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getTimeUnits(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getTimeStep(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getCurrentTime(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getStartTime(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getEndTime(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVarGrid(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVarType(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVarItemSize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVarUnits(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVarNBytes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVarLocation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getValue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getValueAtIndices(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setValue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setValueAtIndices(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridSize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridType(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridRank(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridShape(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridSpacing(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridOrigin(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridX(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridY(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridZ(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridNodeCount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridEdgeCount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridFaceCount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridEdgeNodes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridFaceNodes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridFaceEdges(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGridNodesPerFace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BmiServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'initialize': grpc.unary_unary_rpc_method_handler(
                    servicer.initialize,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.InitializeRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            ),
            'update': grpc.unary_unary_rpc_method_handler(
                    servicer.update,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            ),
            'updateUntil': grpc.unary_unary_rpc_method_handler(
                    servicer.updateUntil,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            ),
            'finalize': grpc.unary_unary_rpc_method_handler(
                    servicer.finalize,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            ),
            'getComponentName': grpc.unary_unary_rpc_method_handler(
                    servicer.getComponentName,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetComponentNameResponse.SerializeToString,
            ),
            'getInputItemCount': grpc.unary_unary_rpc_method_handler(
                    servicer.getInputItemCount,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.SerializeToString,
            ),
            'getOutputItemCount': grpc.unary_unary_rpc_method_handler(
                    servicer.getOutputItemCount,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.SerializeToString,
            ),
            'getInputVarNames': grpc.unary_unary_rpc_method_handler(
                    servicer.getInputVarNames,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarNamesResponse.SerializeToString,
            ),
            'getOutputVarNames': grpc.unary_unary_rpc_method_handler(
                    servicer.getOutputVarNames,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarNamesResponse.SerializeToString,
            ),
            'getTimeUnits': grpc.unary_unary_rpc_method_handler(
                    servicer.getTimeUnits,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetTimeUnitsResponse.SerializeToString,
            ),
            'getTimeStep': grpc.unary_unary_rpc_method_handler(
                    servicer.getTimeStep,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetTimeStepResponse.SerializeToString,
            ),
            'getCurrentTime': grpc.unary_unary_rpc_method_handler(
                    servicer.getCurrentTime,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.SerializeToString,
            ),
            'getStartTime': grpc.unary_unary_rpc_method_handler(
                    servicer.getStartTime,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.SerializeToString,
            ),
            'getEndTime': grpc.unary_unary_rpc_method_handler(
                    servicer.getEndTime,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.Empty.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetTimeResponse.SerializeToString,
            ),
            'getVarGrid': grpc.unary_unary_rpc_method_handler(
                    servicer.getVarGrid,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarGridResponse.SerializeToString,
            ),
            'getVarType': grpc.unary_unary_rpc_method_handler(
                    servicer.getVarType,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarTypeResponse.SerializeToString,
            ),
            'getVarItemSize': grpc.unary_unary_rpc_method_handler(
                    servicer.getVarItemSize,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarItemSizeResponse.SerializeToString,
            ),
            'getVarUnits': grpc.unary_unary_rpc_method_handler(
                    servicer.getVarUnits,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarUnitsResponse.SerializeToString,
            ),
            'getVarNBytes': grpc.unary_unary_rpc_method_handler(
                    servicer.getVarNBytes,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarNBytesResponse.SerializeToString,
            ),
            'getVarLocation': grpc.unary_unary_rpc_method_handler(
                    servicer.getVarLocation,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetVarLocationResponse.SerializeToString,
            ),
            'getValue': grpc.unary_unary_rpc_method_handler(
                    servicer.getValue,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetVarRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetValueResponse.SerializeToString,
            ),
            'getValueAtIndices': grpc.unary_unary_rpc_method_handler(
                    servicer.getValueAtIndices,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GetValueAtIndicesRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetValueAtIndicesResponse.SerializeToString,
            ),
            'setValue': grpc.unary_unary_rpc_method_handler(
                    servicer.setValue,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.SetValueRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            ),
            'setValueAtIndices': grpc.unary_unary_rpc_method_handler(
                    servicer.setValueAtIndices,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.SetValueAtIndicesRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            ),
            'getGridSize': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridSize,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridSizeResponse.SerializeToString,
            ),
            'getGridType': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridType,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridTypeResponse.SerializeToString,
            ),
            'getGridRank': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridRank,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridRankResponse.SerializeToString,
            ),
            'getGridShape': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridShape,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridShapeResponse.SerializeToString,
            ),
            'getGridSpacing': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridSpacing,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridSpacingResponse.SerializeToString,
            ),
            'getGridOrigin': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridOrigin,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridOriginResponse.SerializeToString,
            ),
            'getGridX': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridX,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.SerializeToString,
            ),
            'getGridY': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridY,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.SerializeToString,
            ),
            'getGridZ': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridZ,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.SerializeToString,
            ),
            'getGridNodeCount': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridNodeCount,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.SerializeToString,
            ),
            'getGridEdgeCount': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridEdgeCount,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.SerializeToString,
            ),
            'getGridFaceCount': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridFaceCount,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetCountResponse.SerializeToString,
            ),
            'getGridEdgeNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridEdgeNodes,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridEdgeNodesResponse.SerializeToString,
            ),
            'getGridFaceNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridFaceNodes,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridFaceNodesResponse.SerializeToString,
            ),
            'getGridFaceEdges': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridFaceEdges,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridFaceEdgesResponse.SerializeToString,
            ),
            'getGridNodesPerFace': grpc.unary_unary_rpc_method_handler(
                    servicer.getGridNodesPerFace,
                    request_deserializer=grpc4bmi_dot_bmi__pb2.GridRequest.FromString,
                    response_serializer=grpc4bmi_dot_bmi__pb2.GetGridNodesPerFaceResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'bmi.BmiService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class BmiService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def initialize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/initialize',
            grpc4bmi_dot_bmi__pb2.InitializeRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def update(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/update',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def updateUntil(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/updateUntil',
            grpc4bmi_dot_bmi__pb2.GetTimeResponse.SerializeToString,
            grpc4bmi_dot_bmi__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def finalize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/finalize',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getComponentName(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getComponentName',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetComponentNameResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getInputItemCount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getInputItemCount',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getOutputItemCount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getOutputItemCount',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getInputVarNames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getInputVarNames',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarNamesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getOutputVarNames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getOutputVarNames',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarNamesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getTimeUnits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getTimeUnits',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetTimeUnitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getTimeStep(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getTimeStep',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetTimeStepResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getCurrentTime(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getCurrentTime',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getStartTime(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getStartTime',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getEndTime(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getEndTime',
            grpc4bmi_dot_bmi__pb2.Empty.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetTimeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVarGrid(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getVarGrid',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarGridResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVarType(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getVarType',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarTypeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVarItemSize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getVarItemSize',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarItemSizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVarUnits(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getVarUnits',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarUnitsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVarNBytes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getVarNBytes',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarNBytesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVarLocation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getVarLocation',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetVarLocationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getValue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getValue',
            grpc4bmi_dot_bmi__pb2.GetVarRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetValueResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getValueAtIndices(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getValueAtIndices',
            grpc4bmi_dot_bmi__pb2.GetValueAtIndicesRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetValueAtIndicesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setValue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/setValue',
            grpc4bmi_dot_bmi__pb2.SetValueRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setValueAtIndices(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/setValueAtIndices',
            grpc4bmi_dot_bmi__pb2.SetValueAtIndicesRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridSize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridSize',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridSizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridType(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridType',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridTypeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridRank(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridRank',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridRankResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridShape(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridShape',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridShapeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridSpacing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridSpacing',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridSpacingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridOrigin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridOrigin',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridOriginResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridX(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridX',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridY(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridY',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridZ(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridZ',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridPointsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridNodeCount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridNodeCount',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridEdgeCount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridEdgeCount',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridFaceCount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridFaceCount',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetCountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridEdgeNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridEdgeNodes',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridEdgeNodesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridFaceNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridFaceNodes',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridFaceNodesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridFaceEdges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridFaceEdges',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridFaceEdgesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGridNodesPerFace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/bmi.BmiService/getGridNodesPerFace',
            grpc4bmi_dot_bmi__pb2.GridRequest.SerializeToString,
            grpc4bmi_dot_bmi__pb2.GetGridNodesPerFaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
import os
import subprocess
import time
from os.path import abspath
from tempfile import SpooledTemporaryFile
from typing import Iterable

from packaging.specifiers import SpecifierSet
from packaging.version import Version
from typeguard import typechecked

from grpc4bmi.bmi_grpc_client import BmiClient
from grpc4bmi.exceptions import ApptainerVersionException, DeadContainerException

SUPPORTED_APPTAINER_VERSIONS = '>=1.0.0-rc.2'  # First apptainer release with binaries

def check_apptainer_version_string(version_output: str) -> bool:
    version = version_output.split(' ').pop()
    local_version = Version(version.replace('.el', ''))
    if local_version not in SpecifierSet(SUPPORTED_APPTAINER_VERSIONS):
        raise ApptainerVersionException(f'Unsupported version ({version_output}) of apptainer found, '
                                        f'supported versions {SUPPORTED_APPTAINER_VERSIONS}')
    return True

def check_apptainer_version():
    p = subprocess.Popen(['apptainer', '--version'], stdout=subprocess.PIPE)
    (stdout, _) = p.communicate()
    if p.returncode != 0:
        raise ApptainerVersionException('Unable to determine apptainer version')
    check_apptainer_version_string(stdout.decode('utf-8'))


class BmiClientApptainer(BmiClient):
    """BMI GRPC client for model running inside a `Apptainer <https://apptainer.org/>`_ container.

    On instantization launches an Apptainer container.
    The Apptainer container image is expected to run a BMI GRPC server as its default command.
    The client picks a random port and expects the container to run the BMI GRPC server on that port.
    The port is passed to the container using the BMI_PORT environment variable.

    Args:
        image: Apptainer image.

            For Docker Hub image use `docker://*` or convert it to a Apptainer image file.

            To convert Docker image
            `ewatercycle/walrus-grpc4bmi <https://hub.docker.com/layers/ewatercycle/walrus-grpc4bmi>`_
            with `v0.2.0` tag to `./ewatercycle-walrus-grpc4bmi_v0.2.0.sif` Apptainer image file use:

            .. code-block:: console

              apptainer pull ewatercycle-walrus-grpc4bmi_v0.2.0.sif \\
                docker://ewatercycle/walrus-grpc4bmi:v0.2.0

        input_dirs (Iterable[str]): Directories for input files of model.

            All of directories will be mounted read-only inside Apptainer container on same path as outside container.

        work_dir (str): Working directory for model.

            Directory is mounted inside container and changed into.

            To create a random work directory you could use

            .. code-block:: python

              from tempfile import TemporaryDirectory
              from grpc4bmi.bmi_client_apptainer import BmiClientApptainer

              work_dir = TemporaryDirectory()

              image = 'ewatercycle-walrus-grpc4bmi_v0.2.0.sif'
              client =  BmiClientApptainer(image, work_dir.name)

              # Write config to work_dir and interact with client

              # After checking output in work_dir, clean up
              work_dir.cleanup()

        delay (int): Seconds to wait for Apptainer container to startup, before connecting to it

            Increase when container takes a long time to startup.

        timeout (int): Seconds to wait for gRPC client to connect to server.

            By default will try forever to connect to gRPC server inside container.
            Set to low number to escape endless wait.

        capture_logs (bool): Whether to capture stdout and stderr of container .

            If false then redirects output to null device never to be seen again.
            If true then redirects output to temporary file which can be read with :py:func:`BmiClientApptainer.logs()`.
            The temporary file gets removed when this object is deleted.

    **Example 1: Config file already inside image**

    MARRMoT has an `example config file <https://github.com/wknoben/MARRMoT/blob/master/BMI/Config/BMI_testcase_m01_BuffaloRiver_TN_USA.mat>`_ inside its Docker image.

    .. code-block:: python

        from grpc4bmi.bmi_client_apptainer import BmiClientApptainer
        client = BmiClientApptainer(image='docker://ewatercycle/marrmot-grpc4bmi:latest', 
                                    work_dir='/opt/MARRMoT/BMI/Config')
        client.initialize('/opt/MARRMoT/BMI/Config/BMI_testcase_m01_BuffaloRiver_TN_USA.mat')
        client.update_until(client.get_end_time())
        del client

    **Example 2: Config file in input directory**

    The config file and all other files the model needs are in a directory (`/tmp/input`).
    Use `/tmp/work` to capture any output files like logs generated by model.

    .. code-block:: python

        from grpc4bmi.bmi_client_apptainer import BmiClientApptainer
        # Generate config file called 'config.mat' in `/tmp/input` directory
        client = BmiClientApptainer(image='docker://ewatercycle/marrmot-grpc4bmi:latest',
                                    input_dirs=['/tmp/input'],
                                    work_dir='/tmp/work')
        client.initialize('/tmp/input/config.mat')
        client.update_until(client.get_end_time())
        del client

    **Example 3: Read only input directory with config file in work directory**

    The forcing data is in a shared read-only location like `/shared/forcings/walrus`.
    In the config file (`/tmp/work/walrus.yml`) point to a forcing data file (`/shared/forcings/walrus/PEQ_Hupsel.dat`).

    .. code-block:: python

        from grpc4bmi.bmi_client_apptainer import BmiClientApptainer
        client = BmiClientApptainer(image='ewatercycle-walrus-grpc4bmi_v0.2.0.sif',
                                    input_dirs=['/shared/forcings/walrus'],
                                    work_dir='/tmp/work')
        client.initialize('walrus.yml')
        client.update_until(client.get_end_time())
        del client

    **Example 4: Model writes in sub directory of input directory**

    Some models, for example wflow, write output in a sub-directory of the input directory.
    If the input directory is set with the `input_dirs` argument then the model will be unable to write its output as
    input directories are mounted read-only. That will most likely cause the model to die.
    A workaround is to use the `work_dir` argument with input directory as value instead.
    This will make the whole input directory writable so the model can do its thing.

    When input directory is on a shared disk where you do not have write permission then
    the input dir should be copied to a work directory (`/scratch/wflow`) so model can write.

    .. code-block:: python

        from grpc4bmi.bmi_client_apptainer import BmiClientApptainer
        client = BmiClientApptainer(image='docker://ewatercycle/wflow-grpc4bmi:latest',
                                    work_dir='/scratch/wflow')
        client.initialize('wflow_sbm.ini')
        client.update_until(client.get_end_time())
        del client

    **Example 5: Inputs are in multiple directories**

    A model has its forcings (`/shared/forcings/muese`), parameters (`/shared/model/wflow/staticmaps`)
    and config file (`/tmp/work/wflow_sbm.ini`) in different locations.
    The config file should be set to point to the forcing and parameters files.

    .. code-block:: python

        from grpc4bmi.bmi_client_apptainer import BmiClientApptainer
        client = BmiClientApptainer(image='docker://ewatercycle/wflow-grpc4bmi:latest',
                                    input_dirs=['/shared/forcings/muese',
                                                '/shared/model/wflow/staticmaps'],
                                    work_dir='/tmp/work')
        client.initialize('wflow_sbm.ini')
        client.update_until(client.get_end_time())
        del client

    **Example 6: Run model twice with their own work directory**

    While running 2 or models at the same time you do not want the any config or output to be mixed.

    .. code-block:: python

        from grpc4bmi.bmi_client_apptainer import BmiClientApptainer
        client_muese = BmiClientApptainer(image='docker://ewatercycle/wflow-grpc4bmi:latest',
                                          work_dir='/scratch/wflow-meuse')
        client_muese.initialize('wflow_sbm.meuse.ini')
        client_rhine = BmiClientApptainer(image='docker://ewatercycle/wflow-grpc4bmi:latest',
                                          work_dir='/scratch/wflow-rhine')
        client_rhine.initialize('wflow_sbm.rhine.ini')
        ...
        # Run models and set/get values
        ...
        del client_muese
        del client_rhine

    """

    @typechecked
    def __init__(self, image: str, work_dir: str, input_dirs: Iterable[str] = tuple(), delay=0, timeout=None,
                 capture_logs=True,
                 ):
        if type(input_dirs) == str:
            msg = f'type of argument "input_dirs" must be collections.abc.Iterable; ' \
                  f'got {type(input_dirs)} instead'
            raise TypeError(msg)
        check_apptainer_version()
        host = 'localhost'
        port = BmiClient.get_unique_port(host)
        args = [
            "apptainer",
            "run",
            "--contain",
            "--env", f"BMI_PORT={port}"
        ]

        for raw_input_dir in input_dirs:
            input_dir = abspath(raw_input_dir)
            if not os.path.isdir(input_dir):
                raise NotADirectoryError(input_dir)
            args += ["--bind", f'{input_dir}:{input_dir}:ro']
        self.work_dir = abspath(work_dir)
        if self.work_dir in {abspath(d) for d in input_dirs}:
            raise ValueError('Found work_dir equal to one of the input directories. Please drop that input dir.')
        if not os.path.isdir(self.work_dir):
            raise NotADirectoryError(self.work_dir)
        args += ["--bind", f'{self.work_dir}:{self.work_dir}:rw']
        # Change into working directory
        args += ["--pwd", self.work_dir]
        args.append(image)
        logging.info(f'Running {image} apptainer container on port {port}')
        if capture_logs:
            self.logfile = SpooledTemporaryFile(max_size=2 ** 16,  # keep until 65Kb in memory if bigger write to disk
                                                prefix='grpc4bmi-apptainer-log',
                                                mode='w+t',
                                                encoding='utf8')
            stdout = self.logfile
        else:
            stdout = subprocess.DEVNULL
        self.container = subprocess.Popen(args, preexec_fn=os.setsid, stderr=subprocess.STDOUT, stdout=stdout)
        time.sleep(delay)
        returncode = self.container.poll()
        if returncode is not None:
            raise DeadContainerException(
                f'apptainer container {image} prematurely exited with code {returncode}',
                returncode,
                self.logs()
            )
        super(BmiClientApptainer, self).__init__(BmiClient.create_grpc_channel(port=port, host=host), timeout=timeout)

    def __del__(self):
        if hasattr(self, "container"):
            self.container.terminate()
            self.container.wait()
        if hasattr(self, "logfile"):
            # Force deletion of log file
            self.logfile.close()

    def get_value_ref(self, var_name):
        raise NotImplementedError("Cannot exchange memory references across process boundary")

    def logs(self) -> str:
        """Returns complete combined stdout and stderr written by the Apptainer container.

        When object was created with `log_enable=False` argument then always returns empty string.
        """
        if not hasattr(self, "logfile"):
            return ''

        current_position = self.logfile.tell()
        # Read from start
        self.logfile.seek(0)
        content = self.logfile.read()
        # Write from last position
        self.logfile.seek(current_position)
        return content
